u64 ReadOSTimer(void) {
    LARGE_INTEGER timer;
    QueryPerformanceCounter(&timer);
    return timer.QuadPart;
}

u64 ReadOSTimerFrequency(void) {
    LARGE_INTEGER frequency;
    QueryPerformanceFrequency(&frequency);
    return frequency.QuadPart;
}

// TODO: RDPRU / RDPMC instead on Zen2??? Because AMD has large granularity of 40 or so cycles??? Check CPU Manual!!!!!
u64 ReadCPUTimer(void) {
    __faststorefence();
    u64 result = __rdtsc();
    return (result);
}

u64 ReadFencedCPUTimer(void) {
    unsigned int aux;
    u64 result = __rdtscp(&aux);
    __faststorefence();
    return result;
}

u64 ReadCPUTimerFrequency(u64 wait_time_ms) {
    u64 os_frequency = ReadOSTimerFrequency(); // Counts per second
    u64 os_wait_time = os_frequency * wait_time_ms / 1000;

    u64 cpu_start = ReadCPUTimer();
    u64 os_start = ReadOSTimer();

    u64 os_end;
    u64 os_elapsed = 0;

    while (os_elapsed < os_wait_time) {
        os_end = ReadOSTimer();
        os_elapsed = os_end - os_start;
    }

    u64 cpu_end = ReadCPUTimer();

    u64 cpu_elapsed = cpu_end - cpu_start;
    u64 cpu_frequency = 0;

    if (os_elapsed)
        cpu_frequency = os_frequency * cpu_elapsed / os_elapsed;

    return cpu_frequency;
}

void WarmupProfiler(void) {
    ReadCPUTimer();
    ReadCPUTimer();
    ReadCPUTimer();
    ReadFencedCPUTimer();
}

typedef struct PROFILER_ZONE profiler_zone;
struct PROFILER_ZONE {
    u64 ticks_start;
    u64 ticks_end;
};

profiler_zone StartProfilerZone() {
    profiler_zone result;
    result.ticks_start = ReadCPUTimer();
    result.ticks_end = 0;
    return result;
}

void EndProfilerZone(profiler_zone *zone) {
    zone->ticks_end = ReadFencedCPUTimer();
}

typedef struct PROFILER_RESULT profiler_result;
struct PROFILER_RESULT {
    u32 iterations;
    u64 average_ticks;
    u64 min_ticks;
    u64 max_ticks;
    u64 median_ticks;
    u64 ticks_25;
    u64 ticks_75;
    u64 ticks_99;
};

profiler_result CalculateProfilerResultFromZones(memory_arena *arena, profiler_zone *zones, u32 zone_count) {
    profiler_result result = { 0 };

    result.iterations = zone_count;
    result.min_ticks = U64_MAX;

    u64 ticks;
    u64 total_ticks = 0;

    u64 *sorted_ticks = GetMemoryArenaArray(arena, u64, zone_count);
    for (u32 i = 0; i < zone_count; i++) {
        profiler_zone *zone = &zones[i];
        ticks = zone->ticks_end - zone->ticks_start;

        total_ticks += ticks;

        // find the index to put the data in...
        u32 index = 0;
        for (;index < i; index++) {
            if (sorted_ticks[index] > ticks) {
                break;
            }
        }

        //...move everything from that index one slot away...
        if (index < i) {
            memmove(&sorted_ticks[index + 1], &sorted_ticks[index], sizeof(*sorted_ticks) * i);
        }

        // ...and insert
        sorted_ticks[index] = ticks;
    }

    // Calculate percentiles
    u32 imin = 0;
    u32 i25 = (u32)(0.25 * (f64)zone_count);
    u32 i50 = (u32)(0.5 * (f64)zone_count);
    u32 i75 = (u32)(0.75 * (f64)zone_count);
    u32 i99 = (u32)(0.99 * (f64)zone_count);
    u32 imax = zone_count - 1;
    result.min_ticks = sorted_ticks[imin];
    result.ticks_25 = sorted_ticks[i25];
    result.median_ticks = sorted_ticks[i50];
    result.ticks_75 = sorted_ticks[i75];
    result.ticks_99 = sorted_ticks[i99];
    result.max_ticks = sorted_ticks[imax];

    if (zone_count > 0) {
        result.average_ticks = total_ticks / zone_count;
    }

    ReturnMemoryArenaArray(arena, u64, zone_count);
    return (result);
}
