#ifndef __DANI_LIB_STRING_H
#define __DANI_LIB_STRING_H

// Includes
#include "dani_base.h"
#include "external/stb/stb_sprintf.h"

typedef struct STRING string;
struct STRING {
    u8 *base;
    u32 length;
};

#define STRING_MAX S32_MAX
#define STRING_TYPE_SIZE sizeof(string)

// Initialisers
string String(u8 *base, u32 length);

// Checking functions
b32 IsStringEmpty(string value);
b32 IsStringEqual(string a, string b);
b32 IsStringSame(string a, string b);

// Modification functions
string StringCopy(memory_arena *arena, string value);
string StringConcat(memory_arena *arena, string left, string right);

string StringFormat(memory_arena *arena, const s8 *format, ...);
string StringFormatList(memory_arena *arena, const s8 *format, va_list args);

// Helper macros
#define StringLiteral(literal) String((u8 *)(literal), sizeof(literal) - 1)
#define StringLiteralConst(literal) { .base = ((u8 *)(literal)), .length = (sizeof(literal) - 1) }

#define StringToCString(str) ((char *)(str).base)
#define StringToFormatArgs(str) (str).length, StringToCString(str)

#endif // __DANI_LIB_STRING_H

#ifdef DANI_LIB_STRING_IMPLEMENTATION

#define STB_SPRINTF_IMPLEMENTATION
#include "external/stb/stb_sprintf.h"
#undef STB_SPRINTF_IMPLEMENTATION

string String(u8 *base, u32 length) {
    Assert(S32_MAX >= length);

    string result = { base, length };
    return (result);
}

b32 IsStringEmpty(string value) {
    b32 result = (value.length == 0 || value.base == 0 || value.base[0] == '\0');
    return (result);
}

b32 IsStringEqual(string a, string b) {
    b32 result = B32_TRUE;

    if (a.length != b.length) {
        result = B32_FALSE;
    } else {
        for (u32 len = 0; len < a.length; len++) {
            if (a.base[len] != b.base[len]) {
                result = B32_FALSE;
                break;
            }
        }
    }

    return (result);
}

b32 IsStringSame(string a, string b) {
    b32 result = (a.base == b.base && a.length == b.length);
    return (result);
}

string StringCopy(memory_arena *arena, string value) {
    string result = {0};

    result.base = GetMemoryArenaArray(arena, u8, value.length + 1);
    result.length = value.length;
    
    memcpy(result.base, value.base, value.length);

    result.base[value.length] = 0;

    return (result);
}

string StringConcat(memory_arena *arena, string left, string right) {
    string result = {0};

    u32 length = left.length + right.length;
    result.base = GetMemoryArenaArray(arena, u8, length + 1);
    result.length = length;

    memcpy(result.base, left.base, left.length);
    memcpy(&result.base[left.length], right.base, right.length);

    result.base[result.length] = 0;

    return (result);
}

string StringFormat(memory_arena *arena, const s8 *format, ...) {
    va_list args;
    va_start(args, format);

    string result = StringFormatList(arena, format, args);
    
    va_end(args);
    return (result);
}

string StringFormatList(memory_arena *arena, const s8 *format, va_list args) {
    va_list args_copy;
    va_copy(args_copy, args);

    string result = {0};

    u32 required_length = stbsp_vsnprintf(0, 0, format, args) + 1;
    result.base = GetMemoryArenaArray(arena, u8, required_length);
    result.length = stbsp_vsnprintf(StringToCString(result), required_length, format, args_copy);
    
    result.base[result.length] = 0;

    va_end(args_copy);
    return (result);
}

#endif // DANI_LIB_STRING_IMPLEMENTATION
